/**
 *  Copyright (c) 2014 Keang. All Rights Reserved.
 *
 * The CTICallerID class is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The CTICallerID class is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with the CTICallerID class.  If not, see <http://www.gnu.org/licenses/>.
 */
package uk.co.keang.callerid.driver;

import gnu.io.CommPort;
import gnu.io.CommPortIdentifier;
import gnu.io.NoSuchPortException;
import gnu.io.PortInUseException;
import gnu.io.SerialPort;
import gnu.io.SerialPortEvent;
import gnu.io.SerialPortEventListener;
import gnu.io.UnsupportedCommOperationException;

import java.io.BufferedReader;
import java.io.Closeable;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.PipedInputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Scanner;
import java.util.TooManyListenersException;

import org.apache.log4j.BasicConfigurator;
import org.apache.log4j.Logger;

/**
 * Provides an interface to Crucible Technologies (CTI) Comet USB Caller ID hardware.<br>
 * Once closed the connection can not be re-opened, a new CTICallerID object must be created.<P>
 * Note: The "findcomet.exe" application provided by CTI must be placed in the current working directory
 * if you want to be able to find the port(s) the hardware is attached to<p>
 *
 * Interfacing to a single hardware device is as simple as:
 * <pre>
 * <code>
 *    CTICallerID callerID = new CTICallerID();
 *    callerID.addCallerIDListener(new ICallerIDListener()
 *              {
 *              public void incomingCall(CallerIDEvent event)
 *                  {
 *                  // Handle incoming calls
 *                  }
 *
 *              public void callerIDException(Exception e)
 *                  {
 *                  // Handle any exceptions caught by the driver
 *                  }
 *              });
 * </code>
 * </pre>
 * @author Keang
 * @version 1.0 - 15 Jan 2014
 */
public class CTICallerID implements Closeable
{
private static Logger logger = Logger.getLogger("CallerID");
private final List<ICallerIDListener> callerIDListeners = new ArrayList<ICallerIDListener>();
private SerialPort serialPort;
private String portName;
private volatile boolean runSim;
private Thread simThread;

/**
 * Finds the port(s) CTI Comet USB Caller ID hardware is attached to.
 *
 * @return the ports the hardware is attached to or an empty array if none were found.
 * For a single piece of hardware attached to com port 9 on a windows systems this will be "COM9"
 */
public static String[] findCometPorts()
    {
    ProcessBuilder pb = new ProcessBuilder();
    pb.command("findcomet.exe");

    Process process;
    try
        {
        process = pb.start();
        }
    catch ( IOException exc1 )
        {
        // process failed to start
        exc1.printStackTrace();
        return new String[0];
        }

    BufferedReader reader = new BufferedReader(new InputStreamReader(process.getInputStream()));
    List<String> info = new ArrayList<String>();

    try
        {
        String line = null;
        while ( (line = reader.readLine()) != null )
            {
            line = line.trim();

            String[] ports = line.split(";");

            for ( int i = 0; i < ports.length; i++ )
                {
                info.add(ports[i].trim());
                }
            }
        }
    catch ( IOException exc )
        {
        // A problem occurred reading from findcomet.exe process
        exc.printStackTrace();
        }
    finally
        {
        closeQuietly(process.getErrorStream());
        closeQuietly(process.getInputStream());
        closeQuietly(process.getOutputStream());
        }

    return info.toArray(new String[info.size()]);
    }

/**
 * Force the stream to close quietly
 *
 * @param stream
 */
private static void closeQuietly(OutputStream stream)
    {
    if ( stream == null )
        return;

    try
        {
        stream.close();
        }
    catch ( IOException exc )
        {
        // ignore problems - we are closing quietly
        }
    }

/**
 * Force the stream to close quietly
 *
 * @param stream
 */
private static void closeQuietly(InputStream stream)
    {
    if ( stream == null )
        return;

    try
        {
        stream.close();
        }
    catch ( IOException exc )
        {
        // ignore problems - we are closing quietly
        }
    }

/**
 * Creates a caller ID interface for hardware on the first port that has hardware attached
 *
 * @throws IOException
 * @throws UnsupportedCommOperationException
 * @throws PortInUseException
 * @throws NoSuchPortException if auto find failed to identify any ports
 */
public CTICallerID() throws NoSuchPortException, PortInUseException, UnsupportedCommOperationException, IOException
    {
    String[] ports = findCometPorts();

    if ( ports.length == 0 )
        throw new NoSuchPortException();

    // Open Com port
    connect(ports[0]);
    }

/**
 * Creates a caller ID interface for hardware on the given port
 *
 * @param comPort the com port the hardware is attached to ie COM3
 * @throws IOException
 * @throws UnsupportedCommOperationException
 * @throws PortInUseException
 * @throws NoSuchPortException
 */
public CTICallerID(String comPort) throws NoSuchPortException, PortInUseException, UnsupportedCommOperationException, IOException
    {
    // Open Com port
    connect(comPort);
    }

/**
 * Creates a Caller ID interface running from a phone call simulator.<p>
 * This is used for testing purposes only.
 *
 * @param sim the phone call simulator
 */
public CTICallerID(CallSimulator sim)
    {
    final PipedInputStream is = sim.getInputStream();

    // Open Com port
    final SerialReader reader = new SerialReader(is);
    runSim = true;

    simThread = new Thread(new Runnable()
        {
        @Override
        public void run()
            {
            while ( runSim )
                {
                try
                    {
                    if ( is.available() > 0 )
                        {
                        reader.serialEvent(null);
                        }
                    }
                catch ( Exception exc )
                    {
                    logger.warn("Background caller ID thread error", exc);
                    }

                try
                    {
                    Thread.sleep(50);
                    }
                catch ( InterruptedException exc )
                    {
                    // Do nothing
                    }
                }
            }
        });

    simThread.start();
    }

/**
 * Connects to the com port and sets up the listener
 *
 * @param port the port to connect to
 * @throws NoSuchPortException
 * @throws PortInUseException
 * @throws UnsupportedCommOperationException
 * @throws IOException
 */
private void connect(String port) throws NoSuchPortException, PortInUseException, UnsupportedCommOperationException, IOException
    {
    portName = port;

    CommPortIdentifier portIdentifier = CommPortIdentifier.getPortIdentifier(portName);
    if ( portIdentifier.isCurrentlyOwned() )
        {
        throw new PortInUseException();
        }

    CommPort commPort = portIdentifier.open(this.getClass().getName(), 2000);

    logger.info("CallerID: Opened com port "+commPort);

    if ( commPort instanceof SerialPort )
        {
        serialPort = (SerialPort)commPort;
        serialPort.setSerialPortParams(1200, SerialPort.DATABITS_8, SerialPort.STOPBITS_1, SerialPort.PARITY_NONE);

        try
            {
            serialPort.setFlowControlMode(SerialPort.FLOWCONTROL_NONE);
            }
        catch ( UnsupportedCommOperationException ex )
            {
            System.err.println(ex.getMessage());
            }

        logger.debug("Setup com port "+commPort);

        InputStream in = serialPort.getInputStream();

        try
            {
            serialPort.addEventListener(new SerialReader(in));
            }
        catch ( TooManyListenersException exc )
            {
            // this shouldn't happen as this is the only listener being added
            logger.error("Too many listeners", exc);
            }

        serialPort.notifyOnDataAvailable(true);

        logger.info("CallerID: Initialised");
        }
    else
        {
        logger.error("Error: Only serial ports are suppprted.");

        if ( commPort != null )
            commPort.close();

        throw new UnsupportedCommOperationException("Error: Only serial ports are suppprted. You tried you open port "+portName);
        }
    }

/**
 * Adds an CallerID listener. The listener will be notified of the details all incoming phone calls
 *
 * @param listener the listener to add
 */
public void addCallerIDListener(ICallerIDListener listener)
    {
    callerIDListeners.add(listener);
    }

/**
 * Removes an CallerID listener
 *
 * @param listener the listener to remove
 */
public void removeCallerIDListener(ICallerIDListener listener)
    {
    callerIDListeners.remove(listener);
    }

/**
 * Fires the CallerID listener
 *
 * @param event the event containing the call data
 */
private void fireCallerIDListener(CallerIDEvent event)
    {
    for ( ICallerIDListener listener : callerIDListeners )
        {
        listener.incomingCall(event);
        }
    }

/**
 * Fires the CallerID Exception listener
 *
 * @param e the exception that has been trapped
 */
private void fireCallerIDListener(Exception e)
    {
    for ( ICallerIDListener listener : callerIDListeners )
        {
        listener.callerIDException(e);
        }
    }

/**
 * Closes the serial connection to the CTI hardware and removes all listeners
 */
@Override
public void close()
    {
    // if the simulator is running stop it
    if ( simThread != null )
        {
        runSim = false;
        simThread.interrupt();
        }

    // if a serial port is open close it
    if ( serialPort != null )
        {
        serialPort.removeEventListener();
        serialPort.close();
        }

    // remove all listeners
    callerIDListeners.clear();
    }

//=======================================================================================================
/**
 * Handles the input coming from the serial port. A new line character
 * is treated as the end of a block in this example.
 *
 * This may need to run a background thread to enable detection of end of stream by receiving no data for 100 msec
 */
private class SerialReader implements SerialPortEventListener
    {
    private final InputStream in;
    private final byte[] buffer = new byte[1024];

    public SerialReader(InputStream inp)
        {
        this.in = inp;
        }

    @Override
    public void serialEvent(SerialPortEvent event)
        {
        logger.debug("Serial event received");

        try
            {
            // read the input stream waiting a maximum of 100 mSec
            int len = readInputStreamWithTimeout(in, buffer, 100);

            if ( len == -1 )
                {
                logger.debug("ERROR: Caller ID Input Stream terminated");
                return;
                }

            if ( len == 0 )
                {
                logger.debug("ERROR: Caller ID Input Stream timed out");
                return;
                }

            // copy the data to a new buffer
            byte[] dataArray = new byte[len];
            System.arraycopy(buffer, 0, dataArray, 0, len);

            logger.debug("Caller ID Inp Buf = " + new String(buffer, 0, len));
            logger.debug("Caller ID Inp Buf = " + Arrays.toString(buffer));

            // decode the data and notify listeners
            handleData(dataArray);
            }
        catch ( final IllegalArgumentException e )
            {
            fireCallerIDListener(e);
            }
        catch ( final IOException e )
            {
            logger.error("Failed to read data from hardware", e);

            // closing down is done on a separate thread or this locks up
            Thread t = new Thread(new Runnable()
                    {
                    @Override
                    public void run()
                        {
                        serialPort.removeEventListener();
                        serialPort.close();

                        fireCallerIDListener(e);
                        }
                    });

            t.start();
            }
        }

    /**
     * Reads the input from a stream until the buffer is full,
     * waiting a max of timeoutMillis between each piece of data being available.
     *
     * @param is the stream to read from
     * @param b the byte array to place the data in
     * @param timeoutMillis the maximum time to wait
     * @return the number of bytes read
     * @throws IOException
     */
    private int readInputStreamWithTimeout(InputStream is, byte[] b, int timeoutMillis) throws IOException
        {
        int bufferOffset = 0;
        long maxTimeMillis = System.currentTimeMillis() + timeoutMillis;

        while ( System.currentTimeMillis() < maxTimeMillis && bufferOffset < b.length )
            {
            int readLength = java.lang.Math.min(is.available(), b.length - bufferOffset);

            // if there is something to read reset the timeout
            if (readLength > 0)
                maxTimeMillis = System.currentTimeMillis() + timeoutMillis;

            int readResult = is.read(b, bufferOffset, readLength);

            // if End of Stream
            if ( readResult == -1 )
                break;

            bufferOffset += readResult;

            // sleep for a while to prevent CPU hogging
            try
                {
                Thread.sleep((timeoutMillis/5)+1);
                }
            catch ( InterruptedException exc )
                {
                // do nothing
                }
            }

        return bufferOffset;
        }

    /**
     * Decode the message and send it to listeners
     *
     * @param dataArray
     */
    private void handleData(byte[] dataArray)
        {
        CallerIDMessage decoder = new CallerIDMessage(dataArray);
        CallerIDEvent event = new CallerIDEvent(decoder);

        fireCallerIDListener(event);
        }

    }

//=============================================================================================================
public static void main(String[] args)
    {
    BasicConfigurator.configure();

    System.out.println("Finding CTI port");

    String[] ports = findCometPorts();

    if ( ports.length == 0 )
        {
        System.out.println("CTI comet not found - Terminating");
        }
    else
        {
        System.out.println("Found CTI on com ports = "+ Arrays.toString(ports));
        System.out.println("Attaching to com port "+ports[0]);

        CTICallerID comet = null;

        try
            {
            comet = new CTICallerID(ports[0]);

            comet.addCallerIDListener(new ICallerIDListener()
                {
                @Override
                public void incomingCall(CallerIDEvent event)
                    {
                    System.out.println("Received caller event: "+event);
                    }

                @Override
                public void callerIDException(Exception e)
                    {
                    System.out.println("Received caller error notification: "+e.getMessage());
                    }
                });

            // wait for input
            System.out.println("=========================================");
            System.out.println("Waiting for the phone to ring... ('q' to quit)");
            @SuppressWarnings("resource")
            Scanner s = new Scanner(System.in);

            while( s.hasNextLine() )
                {
                String inp = s.nextLine();

                if ( inp.equalsIgnoreCase("q") )
                    break;
                }

            System.out.println("Closing connection to comet");
            }
        catch ( Exception exc )
            {
            exc.printStackTrace();
            }
        finally
            {
            if ( comet != null )
                comet.close();
            }
        }
    }


}
