/**
 *  Copyright (c) 2014 Keang. All Rights Reserved.
 *
 * The CallSimulator class is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The CallSimulator class is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with the CallSimulator class.  If not, see <http://www.gnu.org/licenses/>.
 */

package uk.co.keang.callerid.driver;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;
import java.util.Arrays;
import java.util.Calendar;
import java.util.Random;
import java.util.Timer;
import java.util.TimerTask;

/**
 * Simulates incoming phone calls.<p>
 * This will always wait at least 200 mSec between sending a series of call data streams.<p>
 * The simulator is used in system testing and does not require the hardware to be attached. You can
 * create a simulator with a random set of phone numbers or you can give it a list of phone numbers to
 * simulate calls from. ie
 *
 * <pre>
 * <code>
 *    // create an array of phone numbers to call where 'null' is number withheld
 *    String[] nums = new String[] {null, "01482571234", "01482576666", "01482571111", "01482570000", "01482575555", "01482578888"};
 *    CallSimulator sim = new CallSimulator(nums);
 *
 *    callerID = new CTICallerID(sim);
 *    callerID.addCallerIDListener(new ICallerIDListener()
 *              {
 *              public void incomingCall(CallerIDEvent event)
 *                  {
 *                  // Handle incoming calls
 *                  }
 *
 *              public void callerIDException(Exception e)
 *                  {
 *                  // Handle any exceptions caught by the driver
 *                  }
 *              });
 *
 *    // call a random number from the supplied list
 *    sim.callRandomNumber();
 * </code>
 * </pre>
 *
 * @author Keang
 * @version 1.0 - 23 Jan 2014
 */
public class CallSimulator
{
private static final long DELAY = 200L;
private PipedInputStream inp;
private PipedOutputStream out;
private final Random r = new Random();

private final String[] numbers;
private int index;
private long prevTime;
private Timer callSimTimer;

/**
 * A simulator with 10 random phone numbers
 */
public CallSimulator()
    {
    this(10);
    }

/**
 * A simulator with the given number of random phone numbers
 */
public CallSimulator(int n)
    {
    numbers = new String[n];

    // make the numbers
    for ( int i = 0; i < numbers.length; i++ )
        {
        StringBuilder num = new StringBuilder("01527");
        for ( int j = 0; j < 6; j++ )
            {
            num.append(r.nextInt(10));
            }

        numbers[i] = num.toString();
        }

    init();
    }

/**
 * A simulator with the given list of phone numbers
 *
 * @param nums the telephone numbers
 */
public CallSimulator(String[] nums)
    {
    numbers = nums;

    init();
    }

/**
 * A simulator with the given list of phone numbers that automatically calls
 * them in order
 *
 * @param nums the telephone numbers
 * @param initialDelay the time to wait before firing any events
 * @param repeatDelay the time to wait between events
 */
public CallSimulator(String[] nums, long initialDelay, long repeatDelay)
    {
    this(nums);

    TimerTask task = new TimerTask()
            {
            @Override
            public void run()
                {
                callNextNumber();
                }
            };

    callSimTimer = new Timer(true);
    callSimTimer.schedule(task, initialDelay, repeatDelay);
    }


/**
 * A simulator with the given list of phone numbers that automatically runs
 * the task at the specified frequency
 *
 * @param nums the telephone numbers
 * @param initialDelay the time to wait before firing any events
 * @param repeatDelay the time to wait between events
 * @param task the task to run
 */
public CallSimulator(String[] nums, long initialDelay, long repeatDelay, TimerTask task)
    {
    this(nums);

    Timer timer = new Timer(true);
    timer.schedule(task, initialDelay, repeatDelay);
    }

private void init()
    {
    index = 0;

    try
        {
        inp = new PipedInputStream();
        out = new PipedOutputStream(inp);
        }
    catch ( IOException exc )
        {
        exc.printStackTrace();
        }
    }

/**
 * Gets the input stream that will contain the caller ID data
 *
 * @return the input stream
 */
PipedInputStream getInputStream() { return inp; }

/**
 * Simulates a call from the next number in the list
 */
public  void callNextNumber()
    {
    callNumber(numbers[index]);

    synchronized ( numbers )
        {
        index++;
        index = index % numbers.length;
        }
    }

/**
 * Simulates a call from a random number from the list
 */
public void callRandomNumber()
    {
    callNumber(numbers[r.nextInt(numbers.length)]);
    }

/**
 * Simulates a call from the given number
 *
 * @param num the number to call
 */
public void callNumber(String num)
    {
    byte[] b = createByteStream(num);

    sendStream(b);
    }

/**
 * Simulates a call from a number withheld
 *
 */
public void callNumberWithHeld()
    {
    byte[] b = createByteStream(null);

    sendStream(b);
    }

/**
 * Sends a corrupted data stream, ie one with a bad checksum
 */
public void sendBadChecksum()
    {
    byte[] b = createByteStream(numbers[r.nextInt(numbers.length)]);
    b[b.length-1]++;

    sendStream(b);
    }

/**
 * Sends a short data stream, ie one which is missing part of the data
 */
public void sendTruncated()
    {
    byte[] b = createByteStream(numbers[r.nextInt(numbers.length)]);
    byte[] b2 = new byte[b.length/2];

    System.arraycopy(b, 0, b2, 0, b2.length);

    sendStream(b2);
    }


/**
 * Sends the data stream
 *
 * @param b the data
 */
private synchronized void sendStream(byte[] b)
    {
    // make sure there is a delay between data streams
    long time = System.currentTimeMillis();

    while ( prevTime + DELAY > time )
        {
        try
            {
            Thread.sleep(prevTime+DELAY-time);
            }
        catch ( InterruptedException exc )
            {
            exc.printStackTrace();
            }

        time = System.currentTimeMillis();
        }

    // send the stream
    for ( int i = 0; i < b.length; i++ )
        {
        // add a random delay of 1 to 30 mSec between each byte sent
        try
            {
            Thread.sleep((long)(Math.random() * 30)+1);
            }
        catch ( InterruptedException exc )
            {
            // do nothing
            }

        try
            {
            out.write(b[i]);
            out.flush();
            }
        catch ( IOException exc )
            {
            exc.printStackTrace();
            break;
            }
        }

    // record the time
    prevTime = System.currentTimeMillis();
    }

private byte[] createByteStream(String num)
    {
    Calendar cal = Calendar.getInstance();

    ByteArrayOutputStream b = new ByteArrayOutputStream();

    b.write(CallerIDMessage.CALLER_ID);
    b.write(0); // remaining message length - fill this later
    int count = 0;
    b.write(CallerIDMessage.CALL_TYPE);
    b.write(1);
    b.write(CallType.VOICE.getCode());
    count += 3;

    b.write(CallerIDMessage.TIMEDATE);
    b.write(8);
    try
        {
        b.write(split(cal.get(Calendar.MONTH)+1, 2));
        b.write(split(cal.get(Calendar.DAY_OF_MONTH), 2));
        b.write(split(cal.get(Calendar.HOUR_OF_DAY), 2));
        b.write(split(cal.get(Calendar.MINUTE), 2));
        }
    catch ( IOException exc )
        {
        exc.printStackTrace();
        }
    count += 10;

    if ( num == null )
        {
        b.write(CallerIDMessage.WHY_NO_NUMBER);
        b.write(1);
        b.write(WhyNoNumber.WITHHELD.getCode());

        count += 3;
        }
    else
        {
        b.write(CallerIDMessage.CALLING_NUMBER);
        b.write(num.length());
        for ( char ch : num.toCharArray() )
            {
            b.write(ch);
            }
        count += 2 + num.length();

        b.write(CallerIDMessage.CALLING_NAME);
        b.write(0);
        count += 2;
        }

    b.write(0); // checksum

    // fill the message length
    byte[] bytes = b.toByteArray();
    bytes[1] = (byte)count;

    // fill the checksum
    byte total = 0;
    for ( byte c : bytes )
        {
        total -= c;
        }

    bytes[bytes.length-1] = total;

    return bytes;
    }

private byte[] split(int num, int len)
    {
    byte[] b = new byte[len];
    Arrays.fill(b, (byte)'0');
    int i = len - 1;

    while ( num > 0 && i >= 0 )
        {
        int k = num % 10;
        b[i] = (byte)Character.forDigit(k, 10);

        num = num / 10;
        i--;
        }

    return b;
    }

/**
 * Closes the streams
 */
public void close()
    {
    if ( callSimTimer != null )
        callSimTimer.cancel();

    try
        {
        inp.close();
        }
    catch ( IOException exc1 )
        {
        exc1.printStackTrace();
        }

    try
        {
        out.close();
        }
    catch ( IOException exc )
        {
        exc.printStackTrace();
        }
    }
}
