/**
 *  Copyright (c) 2014 Keang. All Rights Reserved.
 *
 * The CallerIDMessage class is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The CallerIDMessage class is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with the CallerIDMessage class.  If not, see <http://www.gnu.org/licenses/>.
 */

package uk.co.keang.callerid.driver;

import java.util.Calendar;
import java.util.Date;

import org.apache.log4j.Logger;


/**
 * Decodes a byte array into caller id values ie the message
 *
 * @author Keang
 * @version 1.0 - 15 Jan 2014
 */
public class CallerIDMessage
{
private static Logger logger = Logger.getLogger("CallerID");

final static byte CALLER_ID = (byte)0x80;
final static byte TIMEDATE = 0x01;
final static byte CALLING_NUMBER = 0x02;
final static byte WHY_NO_NUMBER = 0x04;
final static byte CALLING_NAME = 0x07;
final static byte CALL_TYPE = 0x11;
final static byte WITHHELD = 0x50;
final static byte UNAVAILABLE = 0x4F;
final static byte VOICE = 0x01;
final static byte RINGBACK = 0x02;
final static byte MSG_WAITING = (byte)0x81;

private String telNumber;
private String name;
private Date dateTime;
private WhyNoNumber noNumber;
private CallType callType;

/**
 * The data to decode
 *
 * @param data the data
 */
public CallerIDMessage(byte[] data)
    {
    decode(data);
    }

/**
 * Decodes the given data
 * @param data the data
 * @throws IllegalArgumentException if the data does not contain valid caller ID data
 */
void decode(byte[] data) throws IllegalArgumentException
    {
    name = null;
    telNumber = null;
    dateTime = null;
    noNumber = WhyNoNumber.UNAVAILABLE;

    checksum(data);

    // Check Caller ID packet id
    if ( data[0] == CALLER_ID )
        {
        byte bytes_remaining = data[1];
        int i = 2;
        // While still bytes to process
        while ( bytes_remaining > 0 )
            {
            int param_length = 0;
            switch ( data[i] )
                {
                case TIMEDATE:
                    bytes_remaining--;
                    i++; // Move to next byte and decrease bytes left

                    param_length = data[i]; // Read in length of this parameter
                    bytes_remaining--;
                    i++; // Move to next byte and decrease bytes left

                    if ( param_length == 8 )
                        {
                        int month = Integer.parseInt(("" + (char)data[i] + (char)data[i+1]).trim());
                        i += 2;
                        int day = Integer.parseInt(("" + (char)data[i] + (char)data[i+1]).trim());
                        i += 2;
                        int hours = Integer.parseInt(("" + (char)data[i] + (char)data[i+1]).trim());
                        i += 2;
                        int mins = Integer.parseInt(("" + (char)data[i] + (char)data[i+1]).trim());
                        i += 2;

                        Calendar cal = Calendar.getInstance();
                        cal.set(Calendar.MONTH, month-1);
                        cal.set(Calendar.DAY_OF_MONTH, day);
                        cal.set(Calendar.HOUR_OF_DAY, hours);
                        cal.set(Calendar.MINUTE, mins);

                        dateTime = cal.getTime();

                        logger.debug("Msg Part: TimeDate = "+dateTime);
                        }
                    else
                        {
                        i += param_length;
                        }
                    bytes_remaining -= param_length;
                    break;

                case CALLING_NUMBER:
                    {
                    bytes_remaining--;
                    i++; // Move to next byte and decrease bytes left
                    param_length = data[i]; // Read in length of this parameter

                    bytes_remaining--;
                    i++; // Move to next byte and decrease bytes left

                    StringBuilder sb = new StringBuilder();
                    for ( ; param_length > 0; param_length-- )
                        {
                        sb.append((char)data[i]);
                        bytes_remaining--;
                        i++; // move to next byte and decrease bytes left
                        }

                    telNumber = sb.toString();
                    noNumber = null;
                    logger.debug("Msg Part: Tel Num = "+telNumber);
                    break;
                    }
                case CALLING_NAME:
                    {
                    bytes_remaining--;
                    i++; // Move to next byte and decrease bytes left

                    param_length = data[i]; // Read in length of this parameter
                    bytes_remaining--;
                    i++; // Move to next byte and decrease bytes left

                    StringBuilder sb = new StringBuilder();
                    for ( ; param_length > 0; param_length-- )
                        {
                        sb.append((char)data[i]);
                        bytes_remaining--;
                        i++; // move to next byte and decrease bytes left
                        }
                    name = sb.toString();
                    logger.debug("Msg Part: Caller Name = "+name);
                    break;
                    }
                case WHY_NO_NUMBER:
                    bytes_remaining--;
                    i++; // Move to next byte and decrease bytes left
                    param_length = data[i]; // Read in length of this parameter

                    bytes_remaining--;
                    i++; // Move to next byte and decrease bytes left

                    if ( param_length == 1 )
                        {
                        noNumber = WhyNoNumber.getEnum(data[i]);
                        }

                    i += param_length;
                    bytes_remaining -= param_length;
                    logger.debug("Msg Part: Why No Num = "+noNumber);
                    break;

                case CALL_TYPE:
                    bytes_remaining--;
                    i++; // Move to next byte and decrease bytes left
                    param_length = data[i]; // Read in length of this parameter

                    bytes_remaining--;
                    i++; // Move to next byte and decrease bytes left

                    if ( param_length == 1 )
                        {
                        callType = CallType.getEnum(data[i]);
                        }

                    bytes_remaining -= param_length;
                    i += param_length;
                    logger.debug("Msg Part: Call type = "+callType);
                    break;

                // DISMISS ANY OTHER PARAMETERS
                default:
                    // the last byte is the checksum
                    // The checksum is calculated doing a negative sum on all bytes except for the final one
                    // or by summing all bytes including the check sum and having the answer == 0

                    logger.debug("Msg Part: Extra unknown parameter");

                    bytes_remaining--;
                    i++; // Move to next byte and decrease bytes left
                    param_length = data[i]; // Read in length of this parameter
                    bytes_remaining--;
                    i++; // Move to next byte and decrease bytes left
                    bytes_remaining -= param_length;
                    i += param_length;
                    break;

                }
            }
        }
    else
        throw new IllegalArgumentException("The data array doesn't contain caller ID info");

    }

/**
 * Checks the total value of all bytes equals zero
 *
 * @param data the data to test
 * @throws IllegalArgumentException if the summed value of all data and the checksum does not equal zero
 */
private void checksum(byte[] data)
    {
    byte sum = 0;

    for ( byte b : data )
        {
        sum += b;
        }

    if ( sum != 0 )
        throw new IllegalArgumentException("The chechsum value is "+sum+", it should be 0");
    }

/**
 * true if the message contains a number
 *
 * @return true if there is a phone number
 */
public boolean hasNumber()
    {
    return telNumber != null;
    }

/**
 * Gets the callers name if there is one
 *
 * @return the callers name or null
 */
public String getName()
    {
    return name;
    }

/**
 * Gets the callers telephone number if there is one
 *
 * @return the callers telephone number or null
 */
public String getNumber()
    {
    return telNumber;
    }

/**
 * Gets the callers date and time
 *
 * @return the callers date and time or null
 */
public Date getDate()
    {
    return dateTime;
    }

/**
 * Gets the reason why there is no phone number
 *
 * @return The reason or null if there is a phone number
 */
public WhyNoNumber getWhyNoNumber()
    {
    return noNumber;
    }

/**
 * Gets the type of call
 *
 * @return the call type
 */
public CallType getCallType()
    {
    return callType;
    }

@Override
public String toString()
    {
    return dateTime+ ": " + telNumber + ", " + callType.name();
    }
}
